from x2paddle.project_convertor.pytorch import models
resnet_pth_urls = models.resnet.model_urls
resnet18_pth = models.resnet18
resnet34_pth = models.resnet34
resnet50_pth = models.resnet50
resnet101_pth = models.resnet101
resnet152_pth = models.resnet152
resnext50_32x4d_pth = models.resnext50_32x4d
resnext101_32x8d_pth = models.resnext101_32x8d
wide_resnet50_2_pth = models.wide_resnet50_2
wide_resnet101_2_pth = models.wide_resnet101_2

vgg_pth_urls = models.vgg.model_urls
vgg11_pth = models.vgg11
vgg11_bn_pth = models.vgg11_bn
vgg13_pth = models.vgg13
vgg13_bn_pth = models.vgg13_bn
vgg16_pth = models.vgg16
vgg16_bn_pth = models.vgg16_bn
vgg19_pth = models.vgg19
vgg19_bn_pth = models.vgg19_bn
